/*
    $Id: k118.c,v 1.5 2000/09/08 12:07:48 root Exp root $

    k118, kitsrus.com kit 118 data logger
    Copyright (C) 2000  James Cameron (quozl@us.netrek.org)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef lint
static char vcid[] = "$Id: k118.c,v 1.5 2000/09/08 12:07:48 root Exp root $";
#endif /* lint */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>

/* port access functions borrowed from prog84 */
int dev_port_fd = -1;

void open_io()
{
  dev_port_fd = open("/dev/port", O_RDWR);
  if(dev_port_fd < 0)
  {
    perror("/dev/port");
    exit(errno);
  }
}

void close_io()
{
  close(dev_port_fd);
  dev_port_fd = -1;
}

void out_byte(int port, unsigned char byte)
{
  off_t s;
  int r;

  s = lseek(dev_port_fd, port, 0);
  if (s < 0) perror("lseek");
  else if (s != port)
    fprintf(stderr, "hmmm: seeking to %d, went to %ld.\n", port, (long)s);

  r = write(dev_port_fd, &byte, 1);
  if (r != 1)
  {
    fprintf(stderr, "hmmm: write returned %d\n", r);
    if (r < 0) perror("write");
  }
}

unsigned char in_byte(int port)
{
  off_t s;
  int r;
  unsigned char ch = 0;

  s = lseek(dev_port_fd, port, 0);
  if (s < 0)
    perror("lseek");
  else if (s != port)
    fprintf(stderr, "hmmm: seeking to %d, went to %ld.\n", port, (long)s);

  r = read(dev_port_fd, &ch, 1);
  if (r != 1)
  {
    fprintf(stderr, "hmmm: read returned %d\n", r);
    if (r < 0) perror("read");
  }

  return ch;
}

/* bit set, clear and test macros */

#define BITSET(port, mask) out_byte(port,in_byte(port)|mask)
#define BITCLR(port, mask) out_byte(port,in_byte(port)&(~mask))
#define BITTST(port, mask) ((in_byte(port)&mask) != 0)

/* 74ls373 octal latch */

/* pin 1, output enable, active low, connected to pin 16, init */
#define OE_PORT 0x37a
#define OE_MASK 0x4
#define OE_LOWER BITCLR(OE_PORT, OE_MASK)
/* #define OE_RAISE BITSET(OE_PORT, OE_MASK) */

/* pin 2, latch enable, active low, connected to pin 1, strobe */
#define LE_PORT 0x37a
#define LE_MASK 0x1
#define LE_LOWER BITCLR(LE_PORT, LE_MASK);
#define LE_RAISE BITSET(LE_PORT, LE_MASK); 
#define LE_PULSE { LE_LOWER; LE_RAISE; }

/* pins 3, 4, 17, 18, data input, connected to pins d0, d1, d2, d3 */
#define OD_PORT 0x378
#define OD_MASK 0xf
#define OD_SHIFT 0
#define OD_WRITE(x) out_byte(OD_PORT, OD_MASK&(x<<OD_SHIFT))

/* pins 6, 9, 12, 15, data output, connected to pins ack, busy, paperend, selectin, aka high nibble 0x379 */
#define ID_PORT 0x379
#define ID_MASK 0xf0
#define ID_SHIFT 4
#define ID_READ (((in_byte(ID_PORT)&ID_MASK)>>ID_SHIFT)^0x8)


/* max186 eight channel analog to digital converter */

/* pin 18, chip select, active low, connected to pin 14, autofd */
#define CS_PORT 0x37a
#define CS_MASK 0x2
#define CS_RAISE BITCLR(CS_PORT, CS_MASK)
#define CS_LOWER BITSET(CS_PORT, CS_MASK)

/* pin 17, data input, active high, connected to pin 3, d1 */
#define DI_PORT 0x378
#define DI_MASK 0x2
#define DI_RAISE BITSET(DI_PORT, DI_MASK)
#define DI_LOWER BITCLR(DI_PORT, DI_MASK)

/* pin 19, serial clock, active high, connected to pin 2, d0 */
#define SCLK_PORT 0x378
#define SCLK_MASK 0x1
#define SCLK_RAISE BITSET(SCLK_PORT, SCLK_MASK)
#define SCLK_LOWER BITCLR(SCLK_PORT, SCLK_MASK)

/* pin 15, data output, active high, connected to pin 15, error */
#define DO_PORT 0x379
#define DO_MASK 0x8
#define DO_READ BITTST(DO_PORT, DO_MASK)

int analog (int chan)
{
  unsigned char i, req = 0x8e;
  int rep = 0;
  int vect[8] = { 0, 4, 1, 5, 2, 6, 3, 7 };

  req = req | (vect[chan] << 4);
  SCLK_LOWER;
  DI_LOWER;
  CS_LOWER;
  for (i=0;i<8;i++) {
    if (req & 0x80) DI_RAISE; else DI_LOWER;
    SCLK_RAISE; SCLK_LOWER; 
    req *= 2;
  }
  CS_RAISE;
  DI_LOWER;
  usleep(10);
  CS_LOWER;
  for(i=0;i<12;i++) {
    SCLK_RAISE; SCLK_LOWER; 
    rep = (rep << 1) | (DO_READ?1:0);
  }
  CS_RAISE;
  
  return rep;
}

#define MAXANALOG 8
#define MAXDIGITAL 4

int arg_analog[MAXANALOG];
int arg_digital[MAXDIGITAL];
int arg_interval;
int arg_number;
int arg_flush;
int arg_time;
int arg_quiet;
int arg_version;
int arg_verbose;

static void copyright()
{
    fprintf(stderr, "Kit 118 Data Logger, 1.0\n"
	    "Copyright (C) 2000  James Cameron (quozl@us.netrek.org)\n"
	    "%s\n\n"
"This program comes with ABSOLUTELY NO WARRANTY; for details see source.\n"
"This is free software, and you are welcome to redistribute it under certain\n"
"conditions; see source for details.\n\n", vcid);
}

static void usage()
{
  copyright();
  
  fprintf(stderr, 
	  "Usage: k118 [options]\n"
	  "\n"
"Options can be\n"
"    -t  --time           include UNIX time as first field in log\n"
"    -i  --interval n     milliseconds between each sample (default 1000)\n"
"    -n  --number n       number of samples to take (default infinite)\n"
"    -f  --flush-time     seconds between flush of stdout\n"
"    -a  --analog n       include analog channel n in sample\n"
"    -r  --digital-in n   include digital input n in sample\n"
"    -d  --digital-out n  set the digital output to mask n\n"
"    -v  --verbose        be verbose\n"
"    -q  --quiet          be quiet, no copyright declaration\n"
"    -V  --version        state version of program\n"
"    -h  --help           display command usage\n"
	  );
}

int main ( int argc, char *argv[] )
{
  int i, number;
  time_t now, flushed;

  /* reset all command line argument variables */
  for(i=0;i<MAXANALOG;i++) arg_analog[i] = 0;
  for(i=0;i<MAXDIGITAL;i++) arg_digital[i] = 0;
  arg_interval = 1000;
  arg_number = 0;
  arg_flush = 0;
  arg_time = 0;
  arg_quiet = 0;
  arg_version = 0;
  arg_verbose = 0;

  /* process command line arguments */
  for(i=1;i<argc;i++) {
    
    if (!strcmp(argv[i], "-t")||!strcmp(argv[i], "--time")) {
      arg_time++;
      continue;
    }

    if (!strcmp(argv[i], "-i")||!strcmp(argv[i], "--interval")) {
      if ( i++ == argc ) break;
      arg_interval = atoi(argv[i]);
      continue;
    }
    
    if (!strcmp(argv[i], "-n")||!strcmp(argv[i], "--number")) {
      if ( i++ == argc ) break;
      arg_number = atoi(argv[i]);
      continue;
    }
    
    if (!strcmp(argv[i], "-f")||!strcmp(argv[i], "--flush-time")) {
      if ( i++ == argc ) break;
      arg_flush = atoi(argv[i]);
      continue;
    }
    
    if (!strcmp(argv[i], "-a")||!strcmp(argv[i], "--analog")) {
      if ( i++ == argc ) break;
      arg_analog[atoi(argv[i])]++;
      continue;
    }
    
    if (!strcmp(argv[i], "-r")||!strcmp(argv[i], "--digital-in")) {
      if ( i++ == argc ) break;
      arg_digital[atoi(argv[i])]++;
      continue;
    }
    
    if (!strcmp(argv[i], "-v")||!strcmp(argv[i], "--verbose")) {
      arg_verbose++;
      continue;
    }
    
    if (!strcmp(argv[i], "-q")||!strcmp(argv[i], "--quiet")) {
      arg_quiet++;
      continue;
    }
    
    if (!strcmp(argv[i], "-V")||!strcmp(argv[i], "--version")) {
      printf("%s\n", vcid);
      arg_version++;
      continue;
    }
    
    if (!strcmp(argv[i], "-h")||!strcmp(argv[i], "--help")) {
      usage();
      exit(1);
    }

    if (!strcmp(argv[i], "-d")||!strcmp(argv[i], "--digital-out")) {
      if ( i++ == argc ) break;
      open_io();
      OD_WRITE(atoi(argv[i]));
      LE_PULSE;
      exit(0);
    }
    
  }
  
  if (!arg_quiet) {
    /* say g'day */
    copyright();
    if (arg_version) exit(0);
  }

  /* set the device into a known state */
  open_io();
  OE_LOWER;
  CS_RAISE;
  LE_RAISE;

  flushed = time(NULL);
  number = 0;

  while(1) {
    int digital;

    now = time(NULL);

    /* time dump */
    if (arg_time) fprintf(stdout, "%ld\t", now);

    /* digital dump */
    digital = ID_READ;
    for(i=0;i<MAXDIGITAL;i++) {
      if (arg_digital[i])
	fprintf(stdout, "%d\t", digital&(1<<i));
    }

    /* analog dump */
    for(i=0;i<MAXANALOG;i++) {
      if (arg_analog[i])
	fprintf(stdout, "%d\t", analog(i));
    }

    /* terminate the output file */
    fprintf(stdout, "\n");

    /* flush output stream if requested */
    if (arg_flush == 0) {
      fflush(stdout);
    } else {
      if ((now - flushed) > arg_flush) {
	fflush(stdout);
	flushed = now;
      }
    }

    /* exit if number of samples collected matches requested */
    number++;
    if (arg_number != 0) {
      if (number >= arg_number) exit(0);
    }

    /* sleep before the next sample */
    usleep(arg_interval*1000);
  }
}
